from NeverFile import NeverFile
import sys,string
import BinaryDataHandler

dataHandler = BinaryDataHandler.BinaryDataHandler()

class GeometryHeader:
    def __init__(self):
        pass

    def fromFile(self,f):
        f.seek(8,1)
        print 'reading name at',hex(f.tell())
        self.name = f.read(64)
        self.name = self.name[:self.name.find('\0')]
        print self.name
        rootPointer = dataHandler.readUIntFile(f)
        self.rootNodeOffset = MDLFile.modelPointerToOffset(rootPointer)
        self.nodeCount = dataHandler.readUIntFile(f)
        f.seek(24,1)
        self.refCount = dataHandler.readUIntFile(f)
        self.type = dataHandler.readUByteFile(f)
        f.seek(3,1)

class Node:
    def __init__(self):
        pass

    def fromFile(self,f,offset):
        f.seek(offset)
        self.headerFromFile(f)
        if self.hasMesh():
            self.meshHeaderFromFile(f)
        
    def headerFromFile(self,f):
        f.seek(24,1)
        self.inheritColour = dataHandler.readUIntFile(f)
        self.nodeNumber = dataHandler.readUIntFile(f)
        self.name = f.read(32)
        self.name = self.name[:self.name.find('\0')]
        print 'read node:',self.name
        self.parentGeomPointer = dataHandler.readUIntFile(f)
        self.parentNodePointer = dataHandler.readUIntFile(f)
        self.childrenArray = MDLFile.arrayFromFile(f)
        self.controllerKeysArray = MDLFile.arrayFromFile(f)
        self.controllerValuesArray = MDLFile.arrayFromFile(f)
        self.flags = dataHandler.readUIntFile(f)
        print hex(self.flags)
        self.readChildren(f)

    def readChildren(self,f):
        f.seek(self.childrenArray.offset)
        childOffsets = dataHandler.readUIntsFile(f,self.childrenArray.size)
        childOffsets = [MDLFile.modelPointerToOffset(p) for p in childOffsets]
        self.children = []
        for o in childOffsets:
            n = Node()
            n.fromFile(f,o)
            self.children.append(n)
        self.childrenArray = None
        
    def meshHeaderFromFile(self,f):
        f.seek(8,1)
        self.faceArray = MDLFile.arrayFromFile(f)
        self.boundingBox = [dataHandler.readFloatsFile(f,3),
                             dataHandler.readFloatsFile(f,3)]
        self.radius = dataHandler.readFloatFile(f)
        self.centreOfMass = [dataHandler.readFloatFile(f),
                             dataHandler.readFloatFile(f),
                             dataHandler.readFloatFile(f)]
        self.diffuseColour = [dataHandler.readFloatFile(f),
                             dataHandler.readFloatFile(f),
                             dataHandler.readFloatFile(f)]
        self.ambientColour = [dataHandler.readFloatFile(f),
                             dataHandler.readFloatFile(f),
                             dataHandler.readFloatFile(f)]
        self.specularColour = [dataHandler.readFloatFile(f),
                             dataHandler.readFloatFile(f),
                             dataHandler.readFloatFile(f)]
        self.shininess = dataHandler.readFloatFile(f)
        self.shadowFlag = dataHandler.readUIntFile(f)
        self.beamingFlag = dataHandler.readUIntFile(f)
        self.renderFlag = dataHandler.readUIntFile(f)
        self.transparencyHint = dataHandler.readUIntFile(f)
        f.seek(4,1)
        self.texture1 = f.read(64)
        self.texture2 = f.read(64)
        self.texture3 = f.read(64)
        self.tileFade = dataHandler.readUIntFile(f)
        
    def hasHeader(self):
        return self.flags & 0x1

    def isLight(self):
        return self.flags & 0x3

    def isEmitter(self):
        return self.flags & 0x5

    def isReference(self):
        return self.flags & 0x11

    def isTriMesh(self):
        return self.flags & 0x21

    def isSkinMesh(self):
        return self.flags & 0x61

    def isAnimMesh(self):
        return self.flags & 0xA1

    def isDanglyMesh(self):
        return self.flags & 0x121

    def isAABBMesh(self):
        return self.flags & 0x221

    def hasAnim(self):
        return self.flags & 0x80
    
    def hasMesh(self):        
        return self.flags & 0x20

    def hasDangly(self):
        return self.flags & 0x120

    def hasSkin(self):
        return self.flags & 0x40
    
    def typeAsString(self):
        sl = []
        if self.hasHeader():
            sl.append("Header")
        if self.isLight():
            sl.append("Light")
        if self.isEmitter():
            sl.append("Emitter")
        if self.isReference():
            sl.append("Reference")
        if self.hasMesh():
            sl.append("Mesh")
        if self.isTriMesh():
            sl.append("Tri")
        if self.isSkinMesh():
            sl.append("Skin")
        if self.isAnimMesh():
            sl.append("Anim")
        if self.isDanglyMesh():
            sl.append("Dangly")
        if self.isAABBMesh():
            sl.append("AABB")
        sl.append(hex(self.flags))
        return string.join(sl,'-')
    
class MDLFile(NeverFile):
    def __init__(self):
        NeverFile.__init__(self)
        self.rawDataOffset = 0
        self.rawDataSize = 0
        self.totalSize = 0

    def arrayFromFile(f):
        class Array:
            pass
        a = Array()
        a.offset = MDLFile.modelPointerToOffset(dataHandler.readUIntFile(f))
        a.size = dataHandler.readUIntFile(f)
        a.entriesAllocated = dataHandler.readUIntFile(f)
        return a

    arrayFromFile = staticmethod(arrayFromFile)
    
    def headerFromFile(self,f):
        firstVal = dataHandler.readUIntFile(f)
        self.rawDataOffset = dataHandler.readUIntFile(f)
        self.rawDataSize = dataHandler.readUIntFile(f)

    def geometryHeaderFromFile(self,f):
        gh = GeometryHeader()
        gh.fromFile(f)
        return gh

    def modelHeaderFromFile(self,f):
        self.mdlGeometryHeader = self.geometryHeaderFromFile(f)
        f.seek(2,1)
        classification = dataHandler.readUByteFile(f)
        if classification == 0x01:
            self.classification = "Effect"
        elif classification == 0x02:
            self.classification = "Tile"
        elif classification == 0x04:
            self.classification = "Character"
        elif classification == 0x08:
            self.classification = "Door"
        else:
            print "unknown model classification:",classification
            self.classification = "None"
        print self.classification
        self.fogged = dataHandler.readUByteFile(f)
        f.seek(4,1)
        self.animationHeaderPointers = self.arrayFromFile(f)
        self.parentPointer = dataHandler.readUIntFile(f)
        print 'reading bbox at',hex(f.tell())
        self.boundingBox = [[dataHandler.readFloatFile(f),
                             dataHandler.readFloatFile(f),
                             dataHandler.readFloatFile(f)],
                            [dataHandler.readFloatFile(f),
                             dataHandler.readFloatFile(f),
                             dataHandler.readFloatFile(f)]]
        print self.boundingBox
        self.radius = dataHandler.readFloatFile(f)
        self.animScale = dataHandler.readFloatFile(f)
        print self.animScale
        self.superModelName = f.read(64)
        self.superModelName = self.superModelName[:self.superModelName.find('\0')]
        print 'supermodel:',self.superModelName
        print 'done at',hex(f.tell()-0xc)

    def nodeFromFile(self,f,pointer):
        n = Node()
        n.fromFile(f,pointer)
        return n

    def modelPointerToOffset(p):
        return p + 0xC
    modelPointerToOffset = staticmethod(modelPointerToOffset)
    
    def fromFile(self,f):
        self.headerFromFile(f)
        self.modelHeaderFromFile(f)
        print self.mdlGeometryHeader.nodeCount,'nodes'
        print self.nodeFromFile(f,self.mdlGeometryHeader.rootNodeOffset)
        #for n in xrange(self.mdlGeometryHeader.nodeCount):
        #    print self.nodeFromFile(f).typeAsString()

if __name__ == '__main__':
    mdl = MDLFile()
    f = open(sys.argv[1],'rb')
    mdl.fromFile(f)
    
